"""Interactive entry point, contains main pypackage runtime logic."""


import os
import sys
import json
import setuptools
from datetime import datetime

from .cmdline import get_options
from .config import get_config
from .configure import set_value_in_config
from .configure import run_interactive_setup
from .constants import HELP
from .constants import META_NAME
from .constants import VERSION
from .context import SetupContext
from .context import ManifestContext
from .guessing import perform_guesswork


def pypackage_setup(setup_py_commands=None, options=None, additional=""):
    """Interactivly configure and run setuptools from static metadata.

    Used as the main entry point for all pypkg-* entry points except for setup

    Args:
        setup_py_commands: a list of string commands to run with setup.py
        options: the Options object from cmdline already initialized or None
        help: string additional infomation to include in help for this command
    """

    options = options or get_options()
    if options.help:
        raise SystemExit(HELP.format(additional="\n".join([
            line.replace("    ", "", 1).rstrip() for
            line in additional.splitlines()
        ])))
    elif options.version:
        raise SystemExit(VERSION)

    config = get_config()
    starting_metadata = config._metadata

    if not options.no_guess:
        perform_guesswork(config, options)

    # enter interactive setup if requested
    if options.interactive:
        run_interactive_setup(config, options)
    elif options.re_classify:
        set_value_in_config("classifiers", config, config._KEYS)

    if starting_metadata != config._metadata and config._metadata:
        banner = "# {} generated by `{}{}{}` at {} UTC\n".format(
            META_NAME,
            os.path.basename(sys.argv[0]),
            " " * int(len(sys.argv) > 1),
            " ".join(sys.argv[1:]),
            datetime.utcnow().isoformat().split(".")[0],
        )
        with open(META_NAME, "w") as openfile:
            openfile.write(banner + json.dumps(config._metadata, indent=4))

    kwargs = config._as_kwargs

    with ManifestContext(config, options):   # write the MANIFEST.in
        with SetupContext(config, options):  # write the setup.py
            if setup_py_commands:
                # use setuptools to build/install/test/whatever us directly
                sys.argv = ["setup.py"]
                sys.argv.extend(setup_py_commands)
                setuptools.setup(**kwargs)
            else:
                if options.interactive:
                    print(" setup.py ".center(40, "~"))
                print(config)



"""Create and inspect python packages built with pypackage."""


import os
import glob
import codecs
import pytest
import tarfile
import zipfile

from pypackage.commands import build


def verify_wheel(expected, zip_path):
    """Asserts the standard wheel files and `expected` are in the zip."""

    wheel_zip = zipfile.ZipFile(glob.glob(os.path.join(zip_path, "*.whl"))[0])
    wheel_files = [zipf.filename for zipf in wheel_zip.infolist()]

    for expected_file in expected:
        for wheel_file in wheel_files:
            if wheel_file.endswith(expected_file):
                break
        else:
            pytest.fail("{} not found in wheel!".format(expected_file))

    standard_wheel_files = ["DESCRIPTION.rst", "WHEEL", "METADATA", "RECORD",
                            "metadata.json", "top_level.txt"]
    assert len(wheel_files) == len(expected) + len(standard_wheel_files)

    for std_file in standard_wheel_files:
        for file_ in wheel_files:
            if os.path.sep in file_:
                if std_file == os.path.split(file_)[1]:
                    break
        else:
            pytest.fail("{} not found in wheel?".format(std_file))


def verify_source(source_files, gz_path):
    """Ensure the std python dist files and source_files are in the .tar.gz."""

    f_name = glob.glob(os.path.join(gz_path, "*.tar.gz"))[0]
    with tarfile.open(f_name, "r:gz") as tar_file:
        tar_files = tar_file.getnames()

    pkg_full_name = os.path.basename(
        os.path.dirname(gz_path)
    ).split(".tar.gz")[0]
    egg_name = "{}.egg-info".format(pkg_full_name.rsplit("-")[0])  # fragile..
    source_files.extend([
        "PKG-INFO",
        egg_name,
        os.path.join(egg_name, "dependency_links.txt"),
        os.path.join(egg_name, "PKG-INFO"),
        os.path.join(egg_name, "SOURCES.txt"),
        os.path.join(egg_name, "top_level.txt"),
        "setup.cfg",
        "setup.py",
    ])
    assert len(tar_files) == len(source_files) + 1  # +1 for the base dir
    base_dir_skipped = False
    for tar_file in tar_files:
        assert tar_file.startswith(pkg_full_name)
        if os.path.sep in tar_file:
            tar_file = tar_file[tar_file.index(os.path.sep) + 1:]
            assert tar_file in source_files
        elif not base_dir_skipped:
            base_dir_skipped = True
        else:
            pytest.fail("{} not expected in source dist!".format(tar_file))


def verify_artifacts(mod_dir):
    """Asserts the dist directory exists and has two files in it."""

    dist_dir = os.path.join(mod_dir, "dist")
    assert os.path.isdir(dist_dir)
    assert len(os.listdir(dist_dir)) == 2, "should have wheel and src"
    return dist_dir


def test_simple_module(simple_module):
    """Test a single simple python module."""

    build()

    mod_dir, mod_name = simple_module
    dist_dir = verify_artifacts(mod_dir)

    mod_file = "{}.py".format(mod_name)
    verify_wheel([mod_file], dist_dir)
    verify_source([mod_file], dist_dir)


def test_simple_package(simple_package):
    """Ensure a simple normal python package is correctly built."""

    build()
    dist_dir = verify_artifacts(simple_package)

    package_name = os.path.basename(simple_package)
    expected_files = [
        "{}/__init__.py".format(package_name),
        "{}/my_module.py".format(package_name),
    ]

    verify_wheel(expected_files, dist_dir)
    verify_source(expected_files + [package_name], dist_dir)


def test_with_data(with_data):
    """Ensure the non-python data files are being picked up."""

    build()
    root, pkg_root = with_data
    dist_dir = verify_artifacts(root)

    package_name = os.path.basename(pkg_root)
    expected_files = [
        "{}/data/data_1".format(package_name),
        "{}/__init__.py".format(package_name),
        "{}/my_module.py".format(package_name),
    ]
    expected_in_source = expected_files + [
        "{}/data".format(package_name),  # data folder
        "MANIFEST.in",
        package_name,  # base folder
    ]

    verify_wheel(expected_files, dist_dir)
    verify_source(expected_in_source, dist_dir)


def test_manifest_mixin(with_data):
    """Ensure we can write a partial MANIFEST.in file."""

    root, pkg_root = with_data
    pkg_name = os.path.basename(pkg_root)
    with open(os.path.join(pkg_root, "data", "data_2"), "w") as opendata:
        opendata.write("some data")

    with open(os.path.join(root, "MANIFEST.in"), "w") as openmanifest:
        openmanifest.write("include {}/data/data_2\n".format(pkg_name))

    build()
    dist_dir = verify_artifacts(root)

    tar_fname = glob.glob(os.path.join(dist_dir, "*.tar.gz"))[0]
    with tarfile.open(tar_fname, "r:gz") as tar_file:
        tar_file.getnames()
        manifest = [f for f in tar_file.members if "MANIFEST.in" in f.name][0]
        content = tar_file.extractfile(manifest).read()

        assert [codecs.decode(l, "utf-8") for l in content.splitlines()] == [
            "include {}/data/data_2".format(pkg_name),
            "include {}/data/data_1".format(pkg_name),
        ]


def test_with_scripts(with_scripts):
    """Ensure the executable scripts are being included."""

    package_dir, script_dir = with_scripts
    build()
    dist_dir = verify_artifacts(package_dir)

    package_name = os.path.basename(package_dir)
    expected_in_wheel = [
        "/scripts/my_script",
        "{}/__init__.py".format(package_name),
        "{}/my_module.py".format(package_name),
    ]
    expected_in_source = [
        package_name,  # base folder
        script_dir,
        "{}/my_script".format(script_dir),
        "{}/__init__.py".format(package_name),
        "{}/my_module.py".format(package_name),
    ]

    verify_wheel(expected_in_wheel, dist_dir)
    verify_source(expected_in_source, dist_dir)


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])



"""Context managers to handle the creation and of setuptools required files."""


import os


class SetupContext(object):
    """Context manager for writing and (maybe) removing the setup.py.

    Args::

        config: `pypackage.config.Config` object for the package
        options: `pypackage.Options` object of cmd line flags
    """

    def __init__(self, config, options):
        self.config = config
        self.options = options

    def __enter__(self):
        """Write the setup.py to include it in the build."""

        # dump the config out as a setup.py for back-compat...
        # it's not actually used this build, we pass kwargs directly
        with open("setup.py", "w") as opensetup:
            opensetup.write("{}\n".format(str(self.config)))

        return self

    def __exit__(self, *args):
        """Cleans up the setup.py unless interactive, or -m/-s flags."""

        if not any([self.options.interactive, self.options.metadata,
                    self.options.setup]):
            os.remove("setup.py")


class ManifestContext(object):
    """Context manager to write the MANIFEST.in file if required.

    Args::

        config: `pypackage.config.Config` object for the package
        options: `pypackage.Options` object of cmd line flags
    """

    def __init__(self, config, options):
        self.config = config
        self.options = options
        self._clean = False  # clean up on exit?
        # gather any previously existing entries..
        try:
            with open("MANIFEST.in", "r") as openmanifest:
                self.previously_existing = openmanifest.read().splitlines()
        except:
            self.previously_existing = []

    def __enter__(self):
        """Write the MANIFEST.in file if there are data files in use."""

        if not getattr(self.config, "package_data", None):
            return self  # nothing to do

        add_to_manifest = []
        for _, files in self.config.package_data.items():
            for file_ in files:
                include_line = "include {}".format(file_)
                if include_line not in self.previously_existing:
                    add_to_manifest.append(include_line)

        if add_to_manifest:
            self._clean = True
            with open("MANIFEST.in", "a") as openmanifest:
                openmanifest.write("{}\n".format("\n".join(add_to_manifest)))

        return self

    def __exit__(self, *args):
        if self.previously_existing:
            with open("MANIFEST.in", "w") as openmanifest:
                openmanifest.write("\n".join(self.previously_existing))
        elif self._clean and not (self.options.metadata or self.options.setup):
            os.remove("MANIFEST.in")



"""Tests to cover the logic in the command line entry points."""


import sys
import mock
import pytest

from pypackage import commands


def test_run_tests():
    """Ensure we call tests correctly..."""

    with mock.patch.object(commands, "pypackage_setup") as patched_setup:
        commands.run_tests()

    patched_setup.assert_called_once_with(
        ["test"],
        additional=commands.run_tests.__doc__,
    )


def test_run_develop():
    """Ensure we call setup.py develop correctly..."""

    with mock.patch.object(commands, "pypackage_setup") as patched_setup:
        commands.develop()

    patched_setup.assert_called_once_with(
        ["develop"],
        additional=commands.develop.__doc__,
    )


@pytest.mark.parametrize("flag", ("s", "m"), ids=("setup", "metadata"))
def test_no_build(reset_sys_argv, flag):
    """Ensure build is not called if setup or metadata is requested."""

    sys.argv = ["py-build", "-{}".format(flag)]
    opts = mock.Mock()  # mock object attributes respond as True-ish
    with mock.patch.object(commands, "pypackage_setup") as patched_setup:
        with mock.patch.object(commands, "get_options", return_value=opts):
            commands.build()

    patched_setup.assert_called_once_with(
        None,
        additional=commands.build.__doc__,
        options=opts,
    )


def test_setup_entry(reset_sys_argv):
    """Setup's entry is a bit different."""

    sys.argv = ["py-setup"]
    with mock.patch.object(commands, "get_config") as patched_config:
        commands.setup()
    patched_config.assert_called_once_with()


@pytest.mark.parametrize("flag", ("-h", "--help"))
def test_setup_entry__help(reset_sys_argv, flag):
    """Ensure we get the docstring as exit help if the flag is set."""

    sys.argv = ["py-setup", flag]
    with pytest.raises(SystemExit) as patched_exit:
        commands.setup()

    for line in commands.setup.__doc__.splitlines():
        line = line.strip()
        if line:
            assert line in patched_exit.value.args[0]


@pytest.mark.parametrize("flag", ("-v", "--version"))
def test_setup_entry__version(reset_sys_argv, flag):
    """Ensure the VERSION static is raised with SystemExit when requested."""

    sys.argv = ["py-test", flag]
    with pytest.raises(SystemExit) as patched_exit:
        commands.setup()

    assert patched_exit.value.args[0] == commands.VERSION


def test_setup_entry__location(reset_sys_argv):
    """Any additonal arguments should be treated as a filepath."""

    sys.argv = ["py-test", "somewhere/on/your/fs"]
    with mock.patch.object(commands, "get_config") as patched_config:
        commands.setup()

    patched_config.assert_called_once_with("somewhere/on/your/fs")


def test_setup_entry__multilocations(reset_sys_argv):
    """All additonal arguments should be treated as a filepath."""

    sys.argv = ["py-test", "place1", "place2"]
    with mock.patch.object(commands, "get_config") as patched_config:
        commands.setup()

    patched_config.assert_any_call("place1")
    patched_config.assert_any_call("place2")

    assert patched_config.call_count == 2


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])
